from cast_ai.se.contollers.cloud_controller_svc import CloudController

from cli.models.config import ConfigObject


class DemoConfigObject:
    def __init__(self, cfg: ConfigObject, cloud_controller: CloudController):
        self.demo_count = cfg.app_config["GENERAL"]["DEMO_NODE_COUNT"]
        self.replicas = cfg.app_config["GENERAL"]["DEMO_REPLICAS"]
        ready_nodes = cloud_controller.get_node_count()
        match cfg.app_inputs["cloud"]:
            case "eks":
                self.scale_to = self.demo_count + ready_nodes
            case "gke":
                self.scale_to = self.demo_count + ready_nodes
            case "aks":
                # Either it`s from hibernation and then the expectancy is ready_nodes = 1 or 0
                # or Refresh then it might be higher then 1
                self.scale_to = self.demo_count + ready_nodes if ready_nodes > 1 else self.demo_count



import os
import platform
import sys

from docopt import ParsedOptions

from cli.constants import LOG_DIR, LOG_FILE
from cli.models.config import ConfigObject
import logging


def init(parsed_options: ParsedOptions) -> ConfigObject:
    cfg = ConfigObject(parsed_options)

    log_level = logging.DEBUG if cfg.app_inputs["debug"] else logging.WARNING
    os.makedirs(LOG_DIR, exist_ok=True)
    logging.basicConfig(filename=os.path.join(LOG_DIR, LOG_FILE),
                        format='%(asctime)s - %(levelname)s:%(name)s - %(message)s',
                        level=log_level)

    logger = logging.getLogger(__name__)

    logger.info(f"{'=' * 200}")
    logger.info(f"Starting executing SE CLI Command Sequencer with the following inputs: {sys.argv[1:]}")
    logger.debug(f"{get_system_info()}")
    return cfg


def get_system_info():
    return {'OS': platform.system(), 'OS Version': platform.version(), 'Machine': platform.machine()}



from typing import Callable
from yaspin import yaspin

from cast_ai.se.models.execution_status import ExecutionStatus

from cli.models.config import ConfigObject
from cli.models.timed_text import TimedText


class BaseOrchestrator:
    def __init__(self, cfg: ConfigObject):
        self._cfg = cfg

    @staticmethod
    def spinner_run(msg: str, action: Callable[[], ExecutionStatus]) -> ExecutionStatus:
        with yaspin(text=TimedText(msg)) as spinner:
            try:
                result = action()
                if result.success:
                    spinner.ok("✅ [Success]")
                else:
                    spinner.fail(f"😢 [Failed:{result.error_message}]")
                return result
            except Exception as e:
                spinner.fail(f"💥 [Failed:{str(e)}]")



from models.config import ConfigObject
from orchestrators.base_orch import BaseOrchestrator


class SnapshotOrchestrator(BaseOrchestrator):
    def __init__(self, cfg: ConfigObject):
        super().__init__(cfg)
        # self._cast_ctrl = CastController(cfg.app_config["CAST"]["CASTAI_API_TOKEN"],
        #                                  cfg.app_config["CAST"]["DEFAULT_CLUSTER_ID"],
        #                                  self._log_level)
        # self._kubectl_ctrl = KubectlController(self._log_level)
        #
        # self._initial_nodes = self._cast_ctrl.get_nodes()
        #
        # self._cloud_ctrl = None
        # self._set_cloud_controller()
        self.snapshot_subcommand_mapping = {

        }

    def execute(self) -> None:
        subcommand = self._cfg.app_inputs["snapshot_subcommand"]
        if subcommand in self.snapshot_subcommand_mapping:
            self.snapshot_subcommand_mapping[subcommand]()
        else:
            raise ValueError(f'Invalid option: {subcommand}')



import logging
import time
from datetime import datetime
from typing import Dict

from models.config import ConfigObject
from orchestrators.base_orch import BaseOrchestrator

from cast_ai.se.contollers.castai_controller_svc import CastController


class AuditOrchestrator(BaseOrchestrator):
    def __init__(self, cfg: ConfigObject):
        super().__init__(cfg)
        self._logger = logging.getLogger(__name__)

        self._cast_ctrl = CastController(cfg.app_config["CAST"]["CASTAI_API_TOKEN"], cfg.cid)

        self.audit_subcommand_mapping = {
            "analyze": self.audit_analyze_sequence,
        }

    def execute(self) -> None:
        subcommand = self._cfg.app_inputs["audit_subcommand"]
        if subcommand in self.audit_subcommand_mapping:
            self.audit_subcommand_mapping[subcommand]()
        else:
            raise ValueError(f'Invalid option: {subcommand}')

    def audit_analyze_sequence(self):
        start_time = time.time()
        try:
            audit_logs = self._cast_ctrl.get_audit()
            error_log_items = list(filter(lambda log_item: "error" in log_item.get("event", {}), audit_logs["items"]))
            if error_log_items:
                print(f"\n{'=' * 20}< Found {len(error_log_items)} log items with errors "
                      f"in {len(audit_logs['items'])} total log items >{'=' * 20}")
                for log_item in error_log_items:
                    event_id, event_time, event_type = self.extract_log_item_essentials(log_item)
                    error_details = log_item["event"]["error"]["details"]

                    print(f"| {event_time} | {event_id} | {event_type} | {error_details} |")
            else:
                print(f"{'=' * 20}< No errors found in log items >{'=' * 20}")

            interr_prediction_events = list(filter(lambda log_item: "interruptionPredictionNodesToRebalance" in
                                                                    log_item.get("eventType", {}), audit_logs["items"]))
            if interr_prediction_events:
                print(f"\n{'=' * 20}< Found {len(interr_prediction_events)} interruptions "
                      f"in {len(audit_logs['items'])} total log items >{'=' * 20}")
                for log_item in interr_prediction_events:
                    self.analyze_interruption(audit_logs, log_item)

        except Exception as e:
            self._logger.error(f"Something went wrong:[{str(e)}]")
        print(f"It took {int(time.time() - start_time)} seconds to analyze")

    def analyze_interruption(self, audit_logs, log_item):
        event_id, event_time, event_type = self.extract_log_item_essentials(log_item)
        instance_type = log_item["event"]["instance_type"]
        nodes = log_item["event"]["node_ids"]
        print(f"{event_type} for {instance_type} at {event_time}")
        for node in nodes:
            node_events = list(filter(lambda log_item: node in log_item.get(
                "event", {}).get("node", {}).get("id", {}), audit_logs["items"]))
            storted_node_events = sorted(node_events, key=lambda x: x['time'])
            print(f"\n{'=' * 10}< {node} events found for  >{'=' * 10}")
            time_of_last_event = None
            for event in storted_node_events:
                time_between_events = None
                node_event_time = datetime.strptime(event["time"], '%Y-%m-%dT%H:%M:%S.%fZ')
                if time_of_last_event is not None:
                    time_between_events = node_event_time - time_of_last_event
                node_event_type = event["eventType"]
                node_event_id = event["id"]
                time_between_events_str = "N/A"
                if time_between_events is not None:
                    total_seconds = time_between_events.total_seconds()
                    total_minutes = total_seconds / 60
                    time_between_events_str = (f"The total seconds since the last event was "
                                               f"{str(total_seconds)} ({str(total_minutes)} minutes)")

                print(f"| {node_event_time} | {node_event_id} | {node_event_type} | "
                      f"{time_between_events_str}")
                time_of_last_event = node_event_time

    @staticmethod
    def extract_log_item_essentials(log_item: Dict) -> tuple[str, str, str]:
        event_type = log_item["eventType"]
        event_id = log_item["id"]
        event_time = log_item["time"]
        return event_id, event_time, event_type



import logging
import os
import json
from datetime import datetime

from docopt import ParsedOptions

from cli.config_constants import CONFIG_TEMPLATE, CLI_INPUTS_TEMPLATE
from cli.constants import CONFIG_PATH, SUPPORTED_COMMANDS
from cli.services.validators import is_config_file_valid, is_cluster_id


class ConfigObject:
    def __init__(self, parsed_inputs: ParsedOptions, config_file_path=CONFIG_PATH):
        self.config_file_path = config_file_path
        self._logger = logging.getLogger(__name__)
        self.app_config = None
        self._setup_config()
        self.app_inputs = CLI_INPUTS_TEMPLATE
        self.cid = None
        self.context = None
        self._populate_parsed_inputs_data(parsed_inputs)

    def create_empty_template(self):
        self._logger.info(f"Creating empty {CONFIG_PATH} file")
        with open(self.config_file_path, "w") as config_file:
            json.dump(CONFIG_TEMPLATE, config_file)

    def _populate_parsed_inputs_data(self, parsed_inputs: ParsedOptions):
        self.app_inputs["demo"] = parsed_inputs["demo"]
        self.app_inputs["demo_subcommand"] = parsed_inputs["<prep | off>"]
        self.app_inputs["cloud"] = parsed_inputs["<eks | aks | gcp>"]
        self.app_inputs["audit"] = parsed_inputs["audit"]
        self.app_inputs["audit_subcommand"] = parsed_inputs["<analyze>"]
        self.app_inputs["cluster_id"] = parsed_inputs["--cluster_id"]
        self.app_inputs["help"] = parsed_inputs["--help"]
        self.app_inputs["debug"] = parsed_inputs["--debug"]
        self._set_cid(parsed_inputs)

        for command, value in parsed_inputs.items():
            if command in SUPPORTED_COMMANDS and value:
                self.app_inputs["command"] = command

    def _set_cid(self, parsed_inputs: ParsedOptions):
        cloud = self.app_inputs["cloud"]
        if cloud:
            self._logger.info(f"Cluster id={self.app_config[cloud.upper()]['CLUSTER_ID']} will be used")
            self.cid = self.app_config[cloud.upper()]['CLUSTER_ID']
            self.context = self.app_config[cloud.upper()]['K8S_CONTEXT']
            return

        elif self.app_inputs["cluster_id"]:
            if is_cluster_id(self.app_inputs["cluster_id"]):
                self._logger.info(f"Custom cluster id={self.app_inputs['cluster_id']} will be used")
                self.cid = self.app_inputs["cluster_id"]
                return
            else:
                raise RuntimeError("--cluster_id did not match expected format")
        self.cid = self.app_config['CAST']['DEFAULT_CLUSTER_ID']

    def _setup_config(self):
        if not os.path.exists(self.config_file_path):
            self.create_empty_template()
            self._logger.critical("config.json was missing. Created an empty template.")
            raise RuntimeError("config.json was missing. Created an empty template.")

        with open(self.config_file_path, "r") as config_file:
            configuration_data = json.load(config_file)
        if not is_config_file_valid(configuration_data):
            timestamp = datetime.now().strftime("%m_%d_%H_%M")
            backup_file_path = f"{self.config_file_path}.{timestamp}.json"
            os.rename(self.config_file_path, backup_file_path)
            self._logger.info(f"Config file was renamed to {backup_file_path} and up2date config created")
            self.create_empty_template()
            self._logger.critical("Invalid configuration file structure/version")
            raise RuntimeError("Invalid configuration file structure")
        self.app_config = configuration_data



"""
se-cli: The swiss army command-line for Cast.AI SE`s.

Usage:
    se-cli demo <prep|off> <eks|aks|gcp> [--cluster_id <cluster_id>] [-h|--help] [-d|--debug]
    se-cli snapshot <analyze> [--cluster_id <cluster_id>] [-h|--help] [-d|--debug] (not supported yet)
    se-cli audit <analyze> [--cluster_id <cluster_id>] [-h|--help] [-d|--debug]

Options:
    -h, --help  Show this help message and exit
    -d, --debug  Enable debug logging
    -c, --cluster_id <cluster_id>  (Optional) Specify the cluster ID (will assume correct kubectl context)

Commands:
    demo      Manage the demo environment
    snapshot  Manage snapshots
    audit     Manage audit logs

Subcommands for "demo":
    on       Prep demo environment for demo (Get off hibernation)
    off      Hibernate demo environment
    refresh  Disable BinPacking and prep demo

Optional subcommands for "demo" (overrides -c, --cluster_id <cluster_id> picks cluster_id from config file of cloud)
    eks     Use EKS config cluster_id and context
    aks     Use AKS config cluster_id and context
    gcp     (not supported yet)

"""
import sys

from orchestrators.demo_orch import DemoOrchestrator
from orchestrators.snapshot_orch import SnapshotOrchestrator
from orchestrators.audit_orch import AuditOrchestrator
from services.misc_svc import init

from docopt import docopt


commands_table = {
    "demo": DemoOrchestrator,
    "snapshot": SnapshotOrchestrator,
    "audit": AuditOrchestrator,
}

if __name__ == '__main__':
    try:
        # TODO: check version once pypi is up
        parsed_options = docopt(__doc__, sys.argv[1:])
        cfg = init(parsed_options)
        orchestrator_class = commands_table[cfg.app_inputs["command"]]
        main_orch = orchestrator_class(cfg)
        main_orch.execute()
    except Exception as e:
        print(f"An error occurred: {e}")



import logging
import time

from cast_ai.se.contollers.kubectl_controller_svc import KubectlController
from cast_ai.se.contollers.castai_controller_svc import CastController
from cast_ai.se.contollers.eks_controller_svc import EKSController
from cast_ai.se.contollers.aks_controller_svc import AKSController
from cast_ai.se.contollers.gke_controller_svc import GKEController

from cast_ai.se.models.execution_status import ExecutionStatus
from cast_ai.se.models.cloud_confs import AksConfig, EksConfig, GkeConfig

from cli.orchestrators.base_orch import BaseOrchestrator
from cli.services.validators import validate_kubectl_context
from cli.models.config import ConfigObject
from cli.models.demo_cfg import DemoConfigObject


class DemoOrchestrator(BaseOrchestrator):
    def __init__(self, cfg: ConfigObject):
        super().__init__(cfg)
        self._logger = logging.getLogger(__name__)

        self._cast_ctrl = CastController(cfg.app_config["CAST"]["CASTAI_API_TOKEN"], cfg.cid)
        self._kubectl_ctrl = KubectlController(cfg.context)

        self._initial_nodes = self._cast_ctrl.get_nodes()

        self._cloud_ctrl = None
        self._set_cloud_controller()
        self.demo_subcommand_mapping = {
            "prep": self.demo_prep_sequence,
            "off": self.demo_off_sequence,
        }
        validate_kubectl_context(self._kubectl_ctrl, self._cast_ctrl.cluster['providerType'])

    def execute(self) -> None:
        subcommand = self._cfg.app_inputs["demo_subcommand"]
        if subcommand in self.demo_subcommand_mapping:
            self.demo_subcommand_mapping[subcommand]()
        else:
            raise ValueError(f'Invalid option: {subcommand}')

    def demo_prep_sequence(self):
        start_time = time.time()
        demo_cfg = DemoConfigObject(self._cfg, self._cloud_ctrl)
        self._logger.info(f"{'=' * 80}< Starting demo-PREP sequence >{'=' * 80}")
        initial_nodes = self._cast_ctrl.get_nodes()

        if not self._cast_ctrl.is_downscaler_disabled():
            self.spinner_run("Disabling downscaler policy", lambda: self._cast_ctrl.disable_downscaler_policy())

        self.spinner_run(f"Scaling cloud nodes to {demo_cfg.scale_to}",
                         lambda: self._cloud_ctrl.scale(demo_cfg.scale_to))

        all_demo_nodes_up = self.spinner_run("Waiting for new nodes to be ready",
                                             lambda: self._wait_for_new_nodes(len(initial_nodes["items"]),
                                                                              demo_cfg.demo_count))

        self.spinner_run("Reconciling Cluster", lambda: self._cast_ctrl.reconcile())

        self.spinner_run(f"Scaling all deployments to {demo_cfg.replicas} replicas",
                         lambda: self._kubectl_ctrl.scale_deployments(demo_cfg.replicas))
        if all_demo_nodes_up.success:
            self.spinner_run(f'Deleting initial {len(initial_nodes["items"])} fallback nodes ',
                             lambda: self._cast_ctrl.delete_nodes(initial_nodes))
        else:
            print("❗❕ You may want to delete the nodes from previous state (e.g. Fallback nodes)...")

        print(f"It took {int(time.time() - start_time)} seconds to prep demo")

    def demo_off_sequence(self):
        demo_off_cronjob = self._cfg.app_config["KUBECTL"]["DEMO_OFF_CRONJOB"]
        self._logger.info(f"{'=' * 80}< Starting demo-OFF sequence >{'=' * 80} ")

        msg = "Scaling all deployments in default namespace to 0"
        self.spinner_run(msg, lambda: self._kubectl_ctrl.scale_deployments(0))

        self.spinner_run("Disabling CAST Unscheduled pods policy",
                         lambda: self._cast_ctrl.disable_unscheduled_pods_policy())

        msg = f"Manually triggering {demo_off_cronjob}"
        cronjob_triggered = (
            self.spinner_run(msg, lambda: self._kubectl_ctrl.trigger_cronjob(demo_off_cronjob)))

        if cronjob_triggered.success:
            self.spinner_run("Turning off autoscaling cloud", lambda: self._cloud_ctrl.disable_autoscaler())
        else:
            print("❗❕ You may want to turn off autoscaling manually...")

    def _wait_for_new_nodes(self, initial_node_count: int, demo_count: int) -> ExecutionStatus:
        try:
            self._logger.info(f"{'-' * 70}[ Waiting for new nodes to be ready ]")
            scaling_timeout = self._cfg.app_config["GENERAL"]["SCALING_TIMEOUT"]
            start_time = time.time()
            while True:
                nodes = self._cast_ctrl.get_nodes()
                self._logger.debug(f"Current:{len(nodes['items'])} "
                                   f"Initial:{initial_node_count} "
                                   f"Requested:{demo_count}")
                if len(nodes["items"]) == initial_node_count + demo_count:
                    all_nodes_ready = all(node["state"]["phase"] == "ready" for node in nodes["items"])
                    if all_nodes_ready:
                        self._logger.info(f"Waited {time.time() - start_time} for all new nodes to be ready...")
                        return ExecutionStatus()
                    else:
                        node_states = [node["state"]["phase"] for node in nodes["items"]]
                        node_states_str = ", ".join(node_states)
                        self._logger.debug(f"Node states: ({node_states_str})")
                if time.time() - start_time > scaling_timeout:
                    self._logger.warning(f"Timeout ({scaling_timeout}sec) reached while waiting for new nodes")
                    return ExecutionStatus(f"Timeout ({scaling_timeout}sec)")
                time.sleep(1)
        except Exception as e:
            self._logger.error(f"Something went wrong:[{str(e)}]")
            return ExecutionStatus(f"Something went wrong:[{str(e)}]")

    def _set_cloud_controller(self):
        match (self._cast_ctrl.cluster['providerType']):
            case "eks":
                self._cloud_ctrl = EKSController(EksConfig(self._cfg.app_config["EKS"]))
                self._logger.info("Orchestrator is using EKS as Cloud Controller")
            case "aks":
                self._cloud_ctrl = AKSController(AksConfig(self._cfg.app_config["AKS"]))
                self._logger.info("Orchestrator is using AKS as Cloud Controller")
            case "gke":
                self._cloud_ctrl = GKEController(GkeConfig(self._cfg.app_config["GKE"]))
                self._logger.info("Orchestrator is using GKE as Cloud Controller")
            case _:
                self._logger.error(f"Unsupported cloud {self._cast_ctrl.cluster['providerType']}")
                raise ValueError(f"{self._cast_ctrl.cluster['providerType']}")



CONFIG_TEMPLATE = {
    "CAST": {
        "CASTAI_API_TOKEN": "",
        "DEFAULT_CLUSTER_ID": ""
    },
    "EKS": {
        "K8S_CONTEXT": "",
        "CLUSTER_ID": "",
        "REGION": "",
        "NODE_GROUP": "",
        "CLUSTER_NAME": "",
        "AUTOSCALING_GROUP": "",
        "ACCESS_KEY": "",
        "ACCESS_SECRET_KEY": ""
    },
    "AKS": {
        "K8S_CONTEXT": "",
        "CLUSTER_ID": "",
        "CLIENT_ID": "",
        "CLIENT_SECRET": "",
        "TENANT_ID": "",
        "NODE_POOL": "",
        "SUBSCRIPTION_ID": "",
        "CLUSTER_NAME": "",
        "RESOURCE_GROUP": ""
    },
    "GKE": {
        "K8S_CONTEXT": "",
        "CLUSTER_ID": "",
        "PROJECT_ID": "",
        "ZONE": "us-east1-b",
        "CLUSTER_NAME": "",
        "NODE_POOL": "",
        "JSON_PATH": ""
    },
    "GENERAL": {
        "DEMO_NODE_COUNT": 7,
        "DEMO_REPLICAS": 2,
        "SCALING_TIMEOUT": 180

    },
    "KUBECTL": {
        "DEMO_OFF_CRONJOB": "hibernate-pause"
    },
}

CLI_INPUTS_TEMPLATE = {
    "demo": False,
    "demo_subcommand": "",
    "audit": False,
    "audit_subcommand": "",
    "cluster_id": "",
    "help": False,
    "debug": False
}
